/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#ifndef _NODE_
#define _NODE_

typedef void *		NKey;
typedef void *		NData;
typedef unsigned long	NMeta;
typedef unsigned int	NType;

#define MNODE_KEY_UNKNOWN	0
#define MNODE_KEY_STRING	1
#define MNODE_KEY_INT		2
#define MNODE_DATA_UNKNOWN	0x10000
#define MNODE_DATA_STRING	0x10001
#define MNODE_DATA_NUMBER	0x10002
#define MNODE_DATA_MAVL		0x10004
#define MNODE_DATA_MLIST	0x10008
#define MNODE_DATA_MHASH	0x10010


class Session;
class MAVL;
class MFBlkfile;

/** An MNode is the base of all collections.
 It is not the most efficient way of storing info, but
it offers a lot of flexibility. 
The cost of an MNode is 32 bytes */
class MNode
{
	friend class	Session;
	friend class 	MAVL;

	protected:
	/** The value by which we find a node 	*/
	NKey		m_key;
	/** The data which it contains		*/
	NData		m_data;
#ifdef MNODE_HAS_TYPE
	/** The type of node this is.		*/	
	NType		m_type;
#endif

	/** Used by containers, "prev" value 	*/
	MNode	*	m_prev;
	/** Used by containers, "next" value	*/
	MNode	*	m_next;
	/** Used by containers for record keeping */
	NMeta		m_meta;
	/** Used by containers for record keeping */
	int		m_count;
	/** Used by applications. Not used by library */
	unsigned int	m_prop;

	public:
	MNode(NKey key=NULL, NData data=NULL);
	virtual ~MNode();

	/** Set Data in node */
	void 	DataSet(NData);
	/** Get Data from Node */
	NData 	DataGet();
	/** Set key in node */
	void 	KeySet(NKey);
	/** Get Key from Node */
	NKey 	KeyGet();

#ifdef MNODE_HAS_TYPE
	/** Set the type of node this is. */
	NType 	TypeGet();
	/** Get the type of node this is. */
	void	TypeSet(NType type);
#endif
	/** Set next node */
	void 	NextSet(MNode *);
	/** return next node */
	MNode *	NextGet();
	/** Set prev node */
	void 	PrevSet(MNode *);
	/** Get prev node */
	MNode *	PrevGet();
	/** Set meta data */
	void 	MetaSet(NMeta m);
	/** Get meta data */
	NMeta 	MetaGet(void);
	/** Set user property */
	void PropSet(unsigned int prop);
	/** Get user property */
	unsigned int PropGet(void);
	/** Compare similarly formatted keys */
	virtual int 	cmp(NKey key);
	/** Compare similar types of nodes */
	inline int 	cmp(MNode *cn)
	{ return cmp(cn->KeyGet()); }
	/** return min node */
	virtual MNode * Min();
	/** return max node */
	virtual MNode * Max();

	/** Counter methods */
	inline void CountSet(int c){m_count=c;}
	inline int CountGet(void){return m_count;}
	inline int CountInc(void) {return ++m_count;}
	inline int CountDec(void) {return --m_count;}
	
	// Ugly hacks!!!
	inline int DataIntInc() 	{ return ((int *)&m_data)[0]++;}
	inline int DataIntDec() 	{ return ((int *)&m_data)[0]--;}
	inline int DataIntGet() 	{ return (int)m_data;}
	inline void DataIntSet(int data){ m_data = (NData)data;}
	inline int DataIntAdd(int n)	{ return ((int *)&m_data)[0] += n; }
	inline int DataIntSum(int n)	{ return ((int *)&m_data)[0] -= n; }

	inline int MetaIntInc() 	{ return ((int *)&m_meta)[0]++;}
	inline int MetaIntDec() 	{ return ((int *)&m_meta)[0]--;}
	inline int MetaIntGet() 	{ return (int)m_meta;}
	inline void MetaIntSet(int meta){ m_meta = (NMeta)meta;}

	inline char *KeyGetChar() { return (char *) m_key; };
	inline int scmp(char * key)
	{ 	int c  = ((char *)key)[0] - ((char *)m_key)[0];
		return (c) ? c : strcmp((char *)key, (char *)m_key);
	}
	inline int scmp(MNode *node)
	{ return scmp((char *)(node->m_key)); }

	// Ugly hack DO NOT use on derived classes!
	void * operator new (size_t size,void *key,int cbkey,void *data=0,int cbdata=0);
	// Ugly hack DO NOT use on derived classes!
	void * operator new (size_t size, void *mem);
	// Ugly hack DO NOT use on derived classes!
	void * operator new (size_t size, char *mem);
	void * operator new (size_t size);

};

typedef MNode * PMNode ;

class MIntNode : public MNode
{
	protected:

	public:
	MIntNode(int key=0, NData data=NULL) : MNode(NULL,data)
	{ m_key = (NKey)key;}
	MIntNode(int key, int data) : MNode(NULL,NULL)
	{ m_key = (NKey)key; m_data = (NData)data;}
	/** Compare similarly formatted keys */
	int 	cmp(NKey key);
	int & GetDataRef(void) { int *pn = (int *)&m_data; return *pn; }
};


/** A node of many nodes */
class MNodeMulti : public MNode
{
	protected:
	int	m_nNodes;
	char *	m_strings;

	public:
	MNodeMulti(void);
	~MNodeMulti();

	void * operator new (size_t size, void *key, int cb, int nNodes);
	MNode * MNodeGet(int n);
	void StringsSet(char *str);
	char *StringsGet(void);
};

/** A two dimetinal node */
class MNode2D : public MNode
{
	protected:
	MNode2D * m_Y_Next;
	MNode2D * m_Y_Prev;

	public:	
	virtual MNode * NextYGet();
	virtual MNode * PrevYGet();

	virtual void	NextYSet(MNode2D *mn2d);
	virtual void 	PrevYSet(MNode2D *mn2d); 
};

/** A three dimentional node */
class MNode3D : public MNode2D
{
	protected:
	MNode3D * m_Z_Next;
	MNode3D * m_Z_Prev;

	public:	
	virtual MNode * NextZGet();
	virtual MNode * PrevZGet();

	virtual void	NextZSet(MNode3D *mn3d);
	virtual void 	PrevZSet(MNode3D *mn3d); 
};


typedef Boolean (*EnumFn)(MNode *mn, void *param1, void *param2);
extern Boolean MNodeDataDel(MNode *mn, void *p1, void *p2);

/** All containers of nodes are MNodeHolders */
class MNodeHolder 
{
	static Boolean deletewalker(MNode *mn, void *param1, void *param2);

	protected:
	virtual	char *fixvalue(char *buffer, char *str);

	public:
	virtual ~MNodeHolder();

	/** Creates and Adds a string pair to the container
		returns the created node.
	*/
	virtual	MNode *Add(char *key, char *data=NULL);

	/** Creates and Adds a binary key and string data to a container
		returns the created node 
	*/
	virtual	MNode *Add(void *key, int cbkey, char *data=NULL);

	/** Creates and Adds a string key and binary data pair to
		the container.
		returns the created node.
	*/
	virtual	MNode *Add(char *key, void *data, int cbdata);

	/** Creates and Adds a binary key/data pair to the container
		returns the created node
	*/
	virtual	MNode *Add(void *key, int cbkey, void *data, int cbdata);
	/** Finds a node based on a node.
		returns the node
	*/
	virtual MNode * Find(MNode *mn);

	// Must be completed by all!

	/** Finds a node based on a key within a container.
		returns the node */
	virtual MNode *	Find(NKey key)=0;

	/** Inserts a node into a container.
		returns TRUE on success */
	virtual	Boolean Insert(MNode *mn)=0;

	/** Loads a container with text values, in the form of "name = value "
		from a file. One set per line. 
		returns TRUE for success.
	*/
	virtual Boolean LoadTextValues(char *fname);

	/** Enumerates all nodes in a container. "ascending" is an optional
		parameter because not all containers may order information
		No return value.
	*/
	virtual void 	Enum(EnumFn fn, void *param1=NULL, void *param2=NULL, Boolean ascending=TRUE)=0;

	/** Deletes all the nodes in a container.
	*/
	virtual void 	DeleteAll(EnumFn fn, void *p1, void *p2);

	/** Acceppts a name=value string and adds it */
	virtual	Boolean split(char *buffer);
};

#ifdef DO_MHNODE

typedef unsigned long 	HBLK;

class MHNode : public MNode
{
	protected:
	friend class MFBlkfile;
	HBLK		m_hblk;
	MFBlkfile	*m_blkf;
	public:
	MHNode();
	~MHNode();
	void * operator new (size_t size, MFBlkfile *blkf, void *key,int cbkey,void *data=0,int cbdata=0);
	void * operator new (size_t size, MFBlkfile *blkf, HBLK hblk);
	void * operator new (size_t size);
	void operator delete (void *node);
};

/** All containers of nodes are */
class MHNodeHolder 
{
	protected:
	MFBlkfile *	m_blkf;
	public:
	virtual ~MHNodeHolder();

	/** Creates and Adds a string pair to the container
		returns the created node.
	*/
	virtual	MHNode *Add(char *key, char *data=NULL);

	/** Creates and Adds a binary key and string data to a container
		returns the created node 
	*/
	virtual	MHNode *Add(void *key, int cbkey, char *data=NULL);

	/** Creates and Adds a string key and binary data pair to
		the container.
		returns the created node.
	*/
	virtual	MHNode *Add(char *key, void *data, int cbdata);

	/** Creates and Adds a binary key/data pair to the container
		returns the created node
	*/
	virtual	MHNode *Add(void *key, int cbkey, void *data, int cbdata);
	/** Finds a node based on a node.
		returns the node
	*/
	virtual MHNode * Find(MNode *mn);

	// Must be completed by all!

	/** Finds a node based on a key within a container.
		returns the node */
	virtual MHNode *	Find(NKey key)=0;

	/** Inserts a node into a container.
		returns TRUE on success */
	virtual	Boolean Insert(MHNode *mn)=0;

	/** Loads a container with text values, in the form of "name = value "
		from a file. One set per line. 
		returns TRUE for success.
	*/
	virtual Boolean LoadTextValues(char *fname);

	/** Enumerates all nodes in a container. "ascending" is an optional
		parameter because not all containers may order information
		No return value.
	*/
	virtual void 	Enum(EnumFn fn, void *param1=NULL, void *param2=NULL, Boolean ascending=TRUE)=0;

};
#endif // DO_MHNODE

#endif
