/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#ifndef _MSOCKET_
#define _MSOCKET_

#ifdef WIN32
#include <winsock.h>
typedef int socklen_t;
#define SHUT_RD		0
#define SHUT_WR		1
#define SHUT_RDWR	2
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#ifdef Freebsd
typedef int socklen_t;
#endif
#ifdef DARWIN
typedef int socklen_t ;
#endif

#ifdef CYGWIN
#define NON_REENTRENT_SOCKET	
#endif

#define SOCKET_SHUT_READ 0
#define SOCKET_SHUT_WRITE 1
#define SOCKET_SHUT_RW	2

#define ISSOCKERR(x)           ((x) < 0)
#define ISSOCKTIMEOUT(x)        ((x) == 0)

#define SOCKET_TIMEOUT     	0
#define SOCKET_RETRY		3
#define SOCKET_LINGER_ABORT	(-1)
#define SOCKET_LINGER_OFF	(0)

#define	SOCKET_TEMPLATE		0x01
#define SOCKET_BOUND		0x02
#define SOCKET_LISTEN		0x04
#define SOCKET_ACCEPT		0x08
#define SOCKET_SELECT_READ	0x10
#define SOCKET_SELECT_WRITE	0x20

#define MAX_HBA	16

/** Socket address union. */
typedef union _msock_addr
{
	struct sockaddr		saddr;
	struct sockaddr_in	saddrin;
}msock_addr;

/** representation of an IP address */
typedef union ipv4
{
        unsigned int    raw;
        unsigned char   p[4];
}IPV4;
 
/** representation of an IP address and netmask. */
typedef struct ipv4netmask
{
        IPV4    mask;
        IPV4    net;
}IPV4NM;

#define EMPTY_SOCK	-1
#define SOCK_ERR	-1

/** A class for using TCP sockets */
class MSock : public MFile
{
	protected:
	unsigned	m_flags;
	int		m_sock;
	msock_addr	m_addr;
	int		m_type;
	int		m_proto;
	int		m_fam;
	int		m_port;
#ifdef NON_REENTRENT_SOCKET	
	MUTEX		m_mutex;
#endif

	/** Fill an IPV4 union from a string. */
	static int FillIPV4(IPV4 *ipv4, char *string);
	/** FIll an IPV4 union from an unsigned int */
	void FillIPV4(IPV4 *ipv4, unsigned int addr)
	{ ipv4->raw = addr; }
	/** Fill an IPV4NM structure from a string. */
	static int FillIPV4NM(IPV4NM *ipv4nm, char *string);
	/** Check an ip address against a string */
	static int checkip(IPV4NM *ip, char *string);
	/** check an IP/netmask against a string */
	static int checkip(IPV4NM *ip, unsigned int bits);

	/** create a socket object from a file descriptor. */
	MSock(int fd, struct sockaddr *psaddr);
	/** change an existing socket to a new file descriptor. */
	void Renew(int fd, struct sockaddr *psaddr);

	public:
	/** create a new socket object. */
	MSock(int port, int fam=AF_INET, int type=SOCK_STREAM, int protocol=0);
	/** Generic socket object. */
	MSock(void);
	/** destroy the socket object */
	virtual ~MSock();
	/** close the file descriptor for a socket. */
	void	Close(void);
	/** Connect with a socket. */
	Boolean	Connect(void);
	/** Connect to a new host by string. */
	Boolean	Connect(char *szhost);
	/** Connect to a new host by host extent. */
	Boolean	Connect(struct hostent *hent);
	/** Reconnects a closed socket to last host */
	Boolean Reconnect();
	/** Write to a socket. */
	ssize_t	write(void *buf, size_t count);
	/** Read from a socket. */
	ssize_t	read(void *buf, size_t count);
	/** Get a string from a socket. */
	ssize_t	gets(char *buf, size_t count);
	/** Write to a socket and retry on a full state. */
	ssize_t	retrywrite(void *buf, size_t count, int retry=SOCKET_RETRY);
	/** Read from a socket and retry on empty conditions. */
	ssize_t	retryread(void *buf, size_t count, int retry=SOCKET_RETRY);
	/** set file descriptor bit for fd_set. */
	int 	FdSet(fd_set *fdset);
	/** clear file descriptor bit for fd_set. */
	int	FdClr(fd_set *fdset);
	/** test a file descriptor bit in an fd_set. */
	Boolean	FdIsSet(fd_set *fdset);
	/** Shut down a socket */
	int 	Shutdown(int type=2);
	/** Set socket linger characteristics */
	Boolean Linger(Boolean fLinger=TRUE, int seconds=0);
	/** Generic getsockopt */
	Boolean	GetSockOpt(int level, int opt, void *val, socklen_t *cb);
	/** Generic setsockopt */
	Boolean	SetSockOpt(int level, int opt, void *val, socklen_t cb);
	/** poll on a socket. */
	Boolean Poll(int events, int *revents, int to);
	/** set TCP_NODELAY on a socket */
	Boolean TcpNoDelay(Boolean fOff);
	/** get char string of IP address */
	char *	IpAddr(void);
	/** Listin for connections on a socket. */
	Boolean	Listen(int queue);
	/** Accept a new connection on a socket. */
	MSock * Accept(void);
	/** Accept on a socket, and renew on new connection. */
	Boolean Accept(MSock *ms);
	/** Bind a socket to its current settings */
	Boolean Bind(void);
	/** Set internal hostent of socket. */
	Boolean	SetHostEnt(struct hostent *hent);
	/** set internal hostent by host name */
	Boolean SetHostByName(char *hname, struct hostent *hent=NULL);
	/** Reset a socket, close exiting socket, and open new. */
	void	Reset(void);
	/** make copy of this object for use in a different thread. */
	MFile *clone(void);
	/** call select with read fd_set options */
	int ReadSelect(int seconds, int useconds=0);
	/** call select with write fd_set options */
	int WriteSelect(int seconds, int useconds=0);
	/** Check unsigned int of IPV4 TCP/IP address, and compare it to 
		allowed IP addresses 
	*/
	static Boolean checkhba(unsigned int parts);
	/** Add netmask/ip to allowed IPV4 TCP/IP addresses. */
	static int allowhba(char *string);
	inline int GetRawSock(void) { return m_sock; }
	inline void SetRawSock(int sock) { m_sock = sock; }
	inline void SetFlag(unsigned flag) {m_flags |= flag; }
};

char **GetLocalNetAddr(int *type);
unsigned int GetLocalIPV4(void);

#endif
